package at.tugraz.iaik.p2p.commons

import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.delay
import kotlinx.coroutines.isActive
import kotlinx.coroutines.yield
import java.io.Closeable
import java.io.IOException
import java.net.InetSocketAddress
import java.net.SocketAddress
import java.nio.ByteBuffer
import java.nio.channels.DatagramChannel
import java.nio.channels.SelectionKey
import java.nio.channels.Selector
import kotlin.coroutines.coroutineContext


class UDPChannel constructor(
    val sendBuffer: ByteBuffer = ByteBuffer.allocateDirect(8096),
    private val rxdBuffer: ByteBuffer = ByteBuffer.allocateDirect(sendBuffer.capacity())
) :
    Closeable {

    constructor(buffer: Int) : this(ByteBuffer.allocateDirect(buffer))


    data class Packet(val data: ByteBuffer, val src: InetSocketAddress)

    var bindAddr: SocketAddress? = null; private set
    private val chan = DatagramChannel.open().apply { configureBlocking(false) }
    private val sel = Selector.open()
    val isClosed: Boolean get() = !chan.isOpen

    init {
        chan.register(sel, SelectionKey.OP_WRITE)
    }

    fun bind(addr: SocketAddress) {
        bindAddr = addr
        chan.bind(addr)
    }

    suspend fun transmit(addr: SocketAddress) {
        sendBuffer.flip()
        chan.send(sendBuffer, addr)
        while (sel.selectNow() == 0) yield()
        sendBuffer.clear()
        sel.selectedKeys().clear()
    }

    suspend fun receive(): Packet {
        rxdBuffer.clear()
        var addr = chan.receive(rxdBuffer)
        while (addr == null) {
            if (!coroutineContext.isActive) throw CancellationException()
          //  yield()
            delay(1)
            addr = chan.receive(rxdBuffer)
        }
        return Packet(rxdBuffer.apply { flip() }, addr as InetSocketAddress)
    }

    @Throws(IOException::class)
    override fun close() {
        var e: Exception? = null
        try {
            chan.close()
        } catch (t: IOException) {
            e = t;
        }
        try {
            sel.close()
        } catch (t: IOException) {
            e = t;
        }
        if (e != null) throw e
    }
}