package at.tugraz.iaik.p2p.evt

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.id.getIdentifier
import at.tugraz.iaik.p2p.hkad.id.putIdentifier
import java.nio.ByteBuffer
import java.security.interfaces.ECPrivateKey
import javax.xml.bind.DatatypeConverter


fun ByteBuffer.putEvent(evt: Event) {
    putIdentifier(evt.criterion)

    putInt(evt.action.size)
    put(evt.action)

    putInt(evt.prev.size)
    put(evt.prev)

    putInt(evt.digest.size)
    put(evt.digest)

}

fun ByteBuffer.getEvent() =
    Event(getIdentifier(),
        ByteArray(getInt()).apply { this@getEvent.get(this) },
        ByteArray(getInt()).apply { this@getEvent.get(this) },
        ByteArray(getInt()).apply { this@getEvent.get(this) })

data class Event(
    val criterion: Identifier,
    val action: ByteArray,
    val prev: ByteArray,
    val digest: ByteArray
) {
    constructor(criterion: Identifier, action: ByteArray, prev: ByteArray, signingKey: ECPrivateKey) : this(
        criterion, action, prev,
        CommonCrypto.sha3(criterion.keyBytes + action)
    )

    override fun toString()="Event(criterion=$criterion, action=${DatatypeConverter.printHexBinary(action)}, prev=${DatatypeConverter.printHexBinary(prev)}, digest=${DatatypeConverter.printHexBinary(digest)})"
}


fun ByteBuffer.getAction(): Action {
    val type = int
    return Action.actionMap[type]?.invoke(this) ?: UnknownAction()
}

fun ByteBuffer.putAction(a: Action) {
    putInt(a.type)
    a.writeTo(this)
}

abstract class Action(val type: Int, deser: (buffer: ByteBuffer) -> Action) {
    init {
        actionMap.putIfAbsent(type, deser)
    }

    companion object {
        val actionMap = mutableMapOf<Int, (buffer: ByteBuffer) -> Action>()
    }

    abstract fun writeTo(to: ByteBuffer)
}

class UnknownAction : Action(0, { UnknownAction() }) {
    override fun writeTo(to: ByteBuffer) {
        TODO("not implemented")
    }
}
