package at.tugraz.iaik.p2p.evt.msg

import at.tugraz.iaik.p2p.evt.Error
import at.tugraz.iaik.p2p.evt.Event
import at.tugraz.iaik.p2p.evt.getEvent
import at.tugraz.iaik.p2p.evt.putEvent
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.id.getIdentifier
import at.tugraz.iaik.p2p.hkad.id.putIdentifier
import at.tugraz.iaik.p2p.hkad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.hkad.node.Node
import java.nio.ByteBuffer
import java.util.*


class ErrorMessage(seqId: Long, origin: Node, val code: Error, val msg: String) :
    ExtendedMessage(
        0.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer ->
            ErrorMessage(
                rd.seqId,
                rd.origin,
                Error.values()[buffer.short.toUShort().toInt()],
                String(ByteArray(buffer.int).apply { buffer.get(this) })
            )
        }) {
    override fun serialize(buffer: ByteBuffer) {

        buffer.putShort(code.ordinal.toUShort().toShort())
        buffer.putInt(msg.length)
        buffer.put(msg.toByteArray())
    }

    override fun toString(): String {
        return "ErrorMessage(seqId=$seqId, origin=$origin, code=$code, msg=\"$msg\")"
    }
}

class PubMessage(seqId: Long, origin: Node, val evt: Event) :
    ExtendedMessage(
        1.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer -> PubMessage(rd.seqId, rd.origin, buffer.getEvent()) }) {
    override fun serialize(buffer: ByteBuffer) = buffer.putEvent(evt)
}

class PubAckMessage(seqId: Long, origin: Node, val evtDigest: ByteArray) :
    ExtendedMessage(
        2.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer ->
            PubAckMessage(
                rd.seqId,
                rd.origin,
                ByteArray(buffer.int).apply { buffer.get(this) })
        }) {
    override fun serialize(buffer: ByteBuffer) {
        buffer.putInt(evtDigest.size)
        buffer.put(evtDigest)
    }

}


class PeekMessage(seqId: Long, origin: Node, val criterionHash: Identifier) :
    ExtendedMessage(
        3.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer -> PeekMessage(rd.seqId, rd.origin, buffer.getIdentifier()) }) {
    override fun serialize(buffer: ByteBuffer) = buffer.putIdentifier(criterionHash)
}

class PeekResultMessage(seqId: Long, origin: Node, val head: ByteArray) :
    ExtendedMessage(
        4.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer ->
            PeekResultMessage(
                rd.seqId,
                rd.origin,
                ByteArray(buffer.int).apply { buffer.get(this) })
        }) {
    override fun serialize(buffer: ByteBuffer) {
        buffer.putInt(head.size)
        buffer.put(head)
    }
}


class PollMessage(seqId: Long, origin: Node, val criterionHash: Identifier, val sinceDigest: ByteArray) :
    ExtendedMessage(
        5.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer ->
            PollMessage(
                rd.seqId,
                rd.origin,
                buffer.getIdentifier(),
                ByteArray(buffer.int).apply { buffer.get(this) })
        }) {
    override fun serialize(buffer: ByteBuffer) {
        buffer.putIdentifier(criterionHash)
        buffer.putInt(sinceDigest.size)
        buffer.put(sinceDigest)
    }
}

class PollResultMessage(seqId: Long, origin: Node, val evtChain: Array<Event>) :
    ExtendedMessage(
        6.toUShort(),
        seqId,
        origin,
        deser = { rd, buffer ->
            PollResultMessage(
                rd.seqId,
                rd.origin,
                Array(buffer.int) { buffer.getEvent() })
        }) {
    override fun serialize(buffer: ByteBuffer) {
        buffer.putInt(evtChain.size)
        evtChain.forEach { buffer.putEvent(it) }
    }


    override fun toString(): String {
        return "PollResultMessage(seqId=$seqId, origin=$origin, evtChain=${Arrays.toString(evtChain)}"
    }

}