package at.tugraz.iaik.p2p.hkad.config

import at.tugraz.iaik.p2p.commons.ECKeyPair
import at.tugraz.iaik.p2p.hkad.id.ECID
import at.tugraz.iaik.p2p.hkad.id.Identifier
import java.net.InetSocketAddress
import java.security.interfaces.ECPrivateKey


private val NONCE_RANGE = 0u.toUShort()

private const val DEFAULT_GET_TO_MS = 5000u

private val NUM_RETRIES = 3u.toUShort()

private val DEFAULT_K = 8u.toUShort()
private val DEFAULT_ALPHA = 3u.toUShort()

private val DEFAULT_DGRAM_SZ = (10u * 1024u).toUShort()


data class Config private constructor(
    val id: Identifier,
    val nonceRange: UShort = NONCE_RANGE,
    val privateKey: ECPrivateKey?,
    val networkTimeoutMs: UInt = DEFAULT_GET_TO_MS,
    val retries: UShort = NUM_RETRIES,
    val k: UShort = DEFAULT_K, val alpha: UShort = DEFAULT_ALPHA,
    val datagramSize: UShort = DEFAULT_DGRAM_SZ,
    val listenAddr: InetSocketAddress
) {
  //  var privateKey: ECPrivateKey?
   //     get() {
    //        val tmp = field
     //       field = null
      //      return tmp
      //  }
       // private set

    val getTimeoutMs: UInt = networkTimeoutMs * retries.toUInt() + 50u

   // init {
     //   this.privateKey = privateKey
   // }

    private constructor(
        nodeId: Identifier,
        nonceRange: UShort = NONCE_RANGE,
        privateKey: ECPrivateKey?,
        networkTimeoutMs: UInt = DEFAULT_GET_TO_MS,
        retries: UShort = NUM_RETRIES,
        kValue: UShort = DEFAULT_K, alpha: UShort = DEFAULT_ALPHA,
        datagramSize: UShort = DEFAULT_DGRAM_SZ,
        listenAddr: String
    ) : this(
        nodeId,
        nonceRange,
        privateKey,
        networkTimeoutMs,
        retries,
        kValue, alpha,
        datagramSize,
        InetSocketAddress(
            listenAddr.substring(0, listenAddr.lastIndexOf(":")),
            Integer.parseInt(listenAddr.substring(listenAddr.lastIndexOf(":") + 1))
        )
    )

    constructor(
        keyPair: ECKeyPair,
        nonceRange: UShort = NONCE_RANGE,
        networkTimeoutMs: UInt = DEFAULT_GET_TO_MS,
        retries: UShort = NUM_RETRIES,
        kValue: UShort = DEFAULT_K, alpha: UShort = DEFAULT_ALPHA,
        datagramSize: UShort = DEFAULT_DGRAM_SZ,
        listenAddr: String
    ) : this(
        ECID(keyPair.public),
        nonceRange,
        keyPair.private,
        networkTimeoutMs,
        retries,
        kValue, alpha,
        datagramSize,
        listenAddr
    )

    constructor(
        nodeId: Identifier,
        nonceRange: UShort = NONCE_RANGE,
        networkTimeoutMs: UInt = DEFAULT_GET_TO_MS,
        retries: UShort = NUM_RETRIES,
        kValue: UShort = DEFAULT_K, alpha: UShort = DEFAULT_ALPHA,
        datagramSize: UShort = DEFAULT_DGRAM_SZ,
        listenAddr: String
    ) : this(
        nodeId,
        nonceRange,
        null,
        networkTimeoutMs,
        retries,
        kValue, alpha,
        datagramSize,
        listenAddr
    )


}
