package at.tugraz.iaik.p2p.hkad.msg

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import at.tugraz.iaik.p2p.commons.sign
import at.tugraz.iaik.p2p.commons.verify
import at.tugraz.iaik.p2p.hkad.ex.CodecException
import at.tugraz.iaik.p2p.hkad.id.ECID
import kotlinx.coroutines.yield
import java.nio.ByteBuffer
import java.security.Signature
import java.security.interfaces.ECPrivateKey
import java.util.concurrent.ConcurrentLinkedQueue


/**
 * Created by Christoph on 21.09.2016.
 */
class Codec(private val privateKey: ECPrivateKey? = null) {

    private val sigQ = ConcurrentLinkedQueue<Signature>()

    init {
        privateKey?.let { for (i in 1..100) sigQ.add(CommonCrypto.createSignature(privateKey)) }

    }

    @Throws(CodecException::class)
    fun decode(buffer: ByteBuffer): Message {
        try {
            val initialPos = buffer.position()
            val msg = buffer.getMessage()
            // msg.origin.lastSeen=System.currentTimeMillis()
            if (msg.origin.id is ECID) {
                // msg.origin.config.id.publicKey ?: throw CryptoException("Origin public key is null")
                try {
                    CommonCrypto.createVerify(msg.origin.id.publicKey).apply {
                        try {
                            val pos = buffer.position()
                            val sigSize = buffer.int
                            val signature = ByteArray(sigSize)
                            buffer.get(signature)
                            buffer.position(initialPos)
                            verify(buffer.slice().apply { limit(pos) }, signature)
                        } finally {
                            //  vrfyQ.put(this)
                        }
                    }

                } catch (t: Throwable) {
                    t.printStackTrace()
                    throw CryptoException(t)
                }
            }
            return msg
        } catch (t: Throwable) {
            throw CodecException(t)
        }
    }

    @Throws(CodecException::class)
    suspend fun encode(msg: Message, dst: ByteBuffer) {
        try {
            val pos = dst.position()
            dst.putMessage(msg)
            if (msg.origin.id is ECID) {
                privateKey ?: throw CryptoException("${msg.origin.id} has no private key!")
                var sig = sigQ.poll()
                while (sig == null) {
                    yield()
                    sig = sigQ.poll()
                }
                sig.apply {
                    val posAfter = dst.position()
                    dst.position(pos)
                    val signature = sign(dst.slice().apply { limit(posAfter) })
                    sigQ.add(this)
                    dst.position(posAfter)
                    dst.putInt(signature.size)
                    dst.put(signature)
                }
            }
        } catch (t: Throwable) {
            throw CodecException(t)
        }
    }
}
