package at.tugraz.iaik.p2p

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.hkad.config.Config
import at.tugraz.iaik.p2p.hkad.id.Identifier
import at.tugraz.iaik.p2p.hkad.id.KadID
import at.tugraz.iaik.p2p.hkad.node.LocalNode
import com.google.common.primitives.UnsignedBytes
import kotlinx.coroutines.*
import mu.KotlinLogging
import org.junit.Test
import java.math.BigInteger
import java.net.BindException
import java.nio.ByteBuffer
import java.util.*
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ForkJoinPool

class Tests {
    val actorCtx1 = ForkJoinPool(1).asCoroutineDispatcher()
    val reactorCtx1 = ForkJoinPool(1).asCoroutineDispatcher()
    val actorCtx2 = ForkJoinPool(1).asCoroutineDispatcher()
    val reactorCtx2 = ForkJoinPool(1).asCoroutineDispatcher()
    @Test
    fun kadID() {

        runBlocking {
            for (i in 10 downTo 1) {
                println(i)
                delay(1000)
            }
        }


        val nx = LocalNode(
            Config(
                nodeId = KadID(ByteArray(Identifier.Type.KAD.numBits / 8)),
                listenAddr = "127.0.0.1:9750"
            ), actorCtx1, reactorCtx1
        )
        nx.config.id.toString()
        val nodes = mutableListOf<LocalNode>()
        for (i in 1..255) {
            val idBis = ByteArray(Identifier.Type.KAD.numBits.toInt() / 8)
            for (h in 0 until idBis.size) {
                idBis[h] = (i ).toByte()
            }
            try {

                val n2 = LocalNode(Config(KadID(idBis), listenAddr = "127.0.0.1:${9750 + i}"), if(i%2==0)actorCtx1 else actorCtx2, if(i%2==0)reactorCtx1 else reactorCtx2)
                n2.config.id.toString()
                nodes.add(n2)
                val job =
                    GlobalScope.launch(if(i%2==0)actorCtx1 else actorCtx2) {
                        n2.bootstrap(nx)
                    }
                runBlocking { job.join() }
            } catch (t: BindException) {
                KotlinLogging.logger("127.0.0.1:${9750 + i}").warn(t) { t.message }
            }

        }

        println("Bootstrapping done ${nx.routingTable}")
        val before = System.currentTimeMillis()
        val stats = ConcurrentHashMap<Identifier, Int>()
        val jobs = mutableListOf<Job>()
        for (i in 1..Math.pow(2.0, 14.0).toInt()) {
            try {

                val j = GlobalScope.launch(if(i%2==0)actorCtx1 else actorCtx2) {
                    val closestForData = nx.put(ByteBuffer.allocate(4).apply { putInt(i) }.array())
                    stats[closestForData.location] = (stats[closestForData.location] ?: 0) + 1
                }
                //runBlocking { j.join() }
                jobs.add(j)
            } catch (t: Throwable) {
                t.printStackTrace()
                System.err.println(t.message)
            }


        }
        runBlocking { jobs.joinAll() }
        val cmp = UnsignedBytes.lexicographicalComparator()
        val duration = System.currentTimeMillis() - before
        val minutes = duration / 60000
        println("Total Time: $minutes:${duration / 1000 - minutes * 60}")



        println("\n\nStatistics:\n=========================     Node     =========================|=#v=|#sel|")
        for ((key, value) in stats.toSortedMap(Comparator { o1, o2 -> cmp.compare(o1.keyBytes, o2.keyBytes) })) {
            println("${key} ${String.format("%4d%5d", 0, value)}")
        }
        jobs.clear()
        nodes.forEach {
            jobs.add(GlobalScope.launch {
                it.close()
            })
        }

        runBlocking { jobs.joinAll() }

    }

    @Test
    fun ecID() {

        runBlocking {
            for (i in 10 downTo 1) {
                println(i)
                delay(1000)
            }
        }


        val nx = LocalNode(
            Config(
                keyPair = CommonCrypto.genKP(ed=false),
                listenAddr = "127.0.0.1:9750"
            ), actorCtx1, reactorCtx1
        )
        nx.config.id.toString()
        val nodes = mutableListOf<LocalNode>()
        for (i in 1..63) {
            try {
                val n2 = LocalNode(Config(keyPair =
                    CommonCrypto.genKP(ed=false), listenAddr = "127.0.0.1:${9750 + i}"), if(i%2==0)actorCtx1 else actorCtx2, if(i%2==0)reactorCtx1 else reactorCtx2)
                n2.config.id.toString()
                nodes.add(n2)
                val job =
                    GlobalScope.launch(if(i%2==0)actorCtx1 else actorCtx2) {
                        n2.bootstrap(nx)
                    }
                runBlocking { job.join() }
            } catch (t: BindException) {
                KotlinLogging.logger("127.0.0.1:${9750 + i}").warn(t) { t.message }
            }

        }

        println("Bootstrapping done ${nx.routingTable}")
        val before = System.currentTimeMillis()
        val stats = ConcurrentHashMap<Identifier, Int>()
        val jobs = mutableListOf<Job>()
        for (i in 1..Math.pow(2.0, 12.0).toInt()) {
            try {

                val j = GlobalScope.launch(if(i%2==0)actorCtx1 else actorCtx2) {
                    val closestForData = nx.put(ByteBuffer.allocate(4).apply { putInt(i) }.array())
                    stats[closestForData.location] = (stats[closestForData.location] ?: 0) + 1
                }
                //runBlocking { j.join() }
                jobs.add(j)
            } catch (t: Throwable) {
                t.printStackTrace()
                System.err.println(t.message)
            }


        }
        runBlocking { jobs.joinAll() }
        val cmp = UnsignedBytes.lexicographicalComparator()
        val duration = System.currentTimeMillis() - before
        val minutes = duration / 60000
        println("Total Time: $minutes:${duration / 1000 - minutes * 60}")



        println("\n\nStatistics:\n=========================     Node     =========================|=#v=|#sel|")
        for ((key, value) in stats.toSortedMap(Comparator { o1, o2 -> cmp.compare(o1.keyBytes, o2.keyBytes) })) {
            println("${key} ${String.format("%4d%5d", 0, value)}")
        }
        jobs.clear()
        nodes.forEach {
            jobs.add(GlobalScope.launch {
                it.close()
            })
        }

        runBlocking { jobs.joinAll() }

    }

}